part of fields.model;

class Bit16Editor extends FieldEditor<int> {
  
  var _expander;
  var _display;
  var _dropdown;
  var _element;
  get element => _element;
  
  Bit16Editor(Field field) : super(field) {
    _element = span(() {
      _expander = toggleButton(() {
        span()..innerHtml = '&#9655;'..classes = ['up'];
        span()..innerHtml = '&#9698;'..classes = ['down'];
      })..onToggle.listen((_) => refreshElement());
      _display = span();
      _dropdown = div(() {
        for (int i = 15; i >= 0; i--) {
          final int bits = 0x0000 | (1 << i);
          final bitField = new Field<bool>('bit$i', () => (value & bits) != 0,
              setter: (bool bit) {
                if (bit) {
                  value |= bits;
                } else {
                  value &= ~bits;
                }
              });
          append(new BoolEditor(bitField).element..style.float = 'left');  
        }        
      })..hidden = true;
    });
  }
  
  void refreshElement() {
    _display.text = '0x${value.toRadixString(16)}';
    _dropdown.hidden = !_expander.isDown;
  }
}

class BoolEditor extends FieldEditor<bool> {
  
  static final _checkBox = new String.fromCharCode(9744);
  static final _check = new String.fromCharCode(9745);
  
  var _element;
  get element => _element;
  
  BoolEditor(Field field) : super(field) {
    final desc = (field.description == null) ? '' : field.description;
    _element = toggleButton(() {
      span()
      ..text = '$_checkBox $desc'
      ..classes = ['up'];
      span()
      ..text = '$_check $desc'
      ..classes = ['down'];
    })..onToggle.listen((bool isDown) {
      if (isEnabled) {
        this.value = isDown;
      }
    });
  }
      
  void refreshElement() {
    _element
    ..disabled = !isEnabled
    ..isDown = value;
  }
}

class ColorEditor extends FieldEditor<String> {
  
  var _picker;
  var _display;
  var _element;
  get element => _element;
  
  ColorEditor(Field field) : super(field) {
    _element = span(() {
      _picker = input()
      ..type = 'color'
      ..style.float = 'left'
      ..onChange.listen((_) {
        _display.value = _picker.value;         
        this.value = _picker.value;
      });
      div(() {
        _display = input()
        ..type = 'text'
        ..disabled = true;
      })
      ..style.overflow = 'hidden';
    });
  }
  
  void refreshElement() {
    _picker
    ..disabled = !isEnabled
    ..value = value;
    _display.value = value;
  }
}

class DateTimeEditor extends FieldEditor<DateTime> {
  
  var _input;
  get element => _input;
  
  DateTimeEditor(Field field) : super(field) {
    _input = input()
    ..type = 'date' // TODO: investigate 'datetime-local'
    ..onChange.listen((_) {
      this.value = _input.valueAsDate;
    });
  }
  
  void refreshElement() {
    _input
    ..disabled = !isEnabled
    ..valueAsDate = value;
  }
}

class DefaultEditor extends FieldEditor<String> {
  
  var _input;
  get element => _input;
    
  DefaultEditor(Field field) : super(field) {
    _input = input()
    ..type = 'text'
    ..onInput.listen((_) {
      try {
        this.value = _input.value;
        _input.classes.remove('error');
      } catch(_) {
        _input.classes.add('error');
      }
    });
  }
  
  void refreshElement() {
    _input.disabled = !isEnabled;
    if (_input.value != value) {
      _input.value = value;
    }
  }
}

class NumberEditor extends FieldEditor<num> {
  
  var _input;
  get element => _input;
  Range _range;
  
  NumberEditor(Field field) : super(field) {
    _input = input()
    ..type = 'number'
    ..onChange.listen((_) {
      final num n = _input.valueAsNumber;
      if (_range != null && !_range.check(n)) {
        _input.valueAsNumber = this.value;
        return;
      }
      this.value = n;
    });
    
    final match = (a) => a is Range;
    if (field.annotations.any(match)) {
      _range = field.annotations.firstWhere(match);
      _range.applyTo(_input);
    }
  }
  
  void refreshElement() {
    _input
    ..disabled = !isEnabled
    ..valueAsNumber = value;
  }
}

class RangeEditor extends FieldEditor<num> {
  
  var _range;
  var _display;
  var _element;
  get element => _element;
  
  RangeEditor(Field field) : super(field) {
    _element = span(() {
      _range = input()
      ..type = 'range'
      ..style.float = 'left'
      ..onChange.listen((_) {
        _display.value = _range.value; 
        this.value = _range.valueAsNumber;
      });
      
      div(() {
        _display = input()
        ..type = 'text'
        ..disabled = true;
      })
      ..style.overflow = 'hidden';
    });

    final match = (a) => a is Range;
    if (field.annotations.any(match)) {
      Range range = field.annotations.firstWhere(match);
      range.applyTo(_range);
    }
  }
  
  void refreshElement() {
    _range
    ..disabled = !isEnabled
    ..valueAsNumber = value;
    _display.value = value.toString();
  }
}

class SelectEditor extends FieldEditor<String> {
  
  Values _values;
  
  var _select;
  get element => _select;   
  
  SelectEditor(Field field) : super(field) {
    _select = select(() {      
      final isValues = (a) => a is Values;
      if (field.annotations.any(isValues)) {
        _values = field.annotations.firstWhere(isValues);        
      }
    })
    ..onChange.listen((_) {
      this.value = _select.value;
    });
  }
  
  void refreshElement() {
    refreshOptions();
    _select
    ..disabled = !isEnabled
    ..value = value;
  }
  
  void refreshOptions() {
    if (_values != null) {
      _select.children.clear();
      for (var v in _values.values) {
        String valueString;
        if (field.transform != null) {
          valueString = field.transform(v);
        } else if (ValueTransform.canTransform(v.runtimeType, String)) {
          valueString = ValueTransform.transform(v, v.runtimeType, String);
        } else {
          valueString = v.toString();
        }
        final option = dom.createElement('option')
        ..text = valueString
        ..value = valueString;
        _select.append(option);
      }
    }
  }
}
