part of fields.model;

final Map<Type, Type> defaultFieldEditors = { 
  bool      : BoolEditor, 
  DateTime  : DateTimeEditor,
  double    : NumberEditor, 
  int       : NumberEditor, 
  num       : NumberEditor
};

final Map<Type, Type> fieldEditors = new Map.from(defaultFieldEditors);

typedef FieldEditor CreateFieldEditor(Field);

final Map<Type, CreateFieldEditor> defaultFieldEditorFactories = {                                                            
  BoolEditor      : (f) => new BoolEditor(f),
  ColorEditor     : (f) => new ColorEditor(f),
  DateTimeEditor  : (f) => new DateTimeEditor(f),
  DefaultEditor   : (f) => new DefaultEditor(f),
  NumberEditor    : (f) => new NumberEditor(f),
  RangeEditor     : (f) => new RangeEditor(f),
  SelectEditor    : (f) => new SelectEditor(f)
};

final Map<Type, CreateFieldEditor> fieldEditorFactories = 
    new Map.from(defaultFieldEditorFactories);

/// An editor for a [Field].
/// 
/// The type parameter `T` sets the [valueType] for this field editor.
abstract class FieldEditor<T> {
  
  static final CreateFieldEditor defaultFactory = (field) {
    Type editorType = field.editorType;
    if (field.editorType == null && fieldEditors.containsKey(field.type)) {
      editorType = fieldEditors[field.type];
    }
    if (editorType == null) {
      editorType = DefaultEditor;
    }
    return fieldEditorFactories[editorType](field);
  };
  
  static CreateFieldEditor factory = defaultFactory;
  
  /// The element rendered by this editor.
  get element;
  
  /// The [Field] edited by this editor.
  final Field field;
  
  /// The type of [value] edited by this editor.
  /// 
  /// This type may differ from the [Field.type] of this editor's [field].  If
  /// they differ, the [value] must be converted to and from this type by this
  /// editor.
  final Type valueType = T;
  
  bool _isEnabled = true;
  /// Whether or not this editor is enabled.
  /// 
  /// This editor is enabled unless:
  /// 
  /// * This is set to `false`.
  /// * This editor's [field] is read-only.
  /// * This editor's [field] cannot be transformed to this editor's 
  /// [valueType].
  bool 
    get isEnabled => 
        _isEnabled &&
        !field.isReadOnly &&
        ( field.transform != null ||
        ValueTransform.canTransform(valueType, field.type) );        
    set isEnabled(bool isEnabled) {
      if (_isEnabled != isEnabled) {
        _isEnabled = isEnabled;
        refreshElement();
      }
    }

  /// The value edited by this editor.
  /// 
  /// This value transforms the [Field.value] of this editor's [field] to be of
  /// [valueType].
  /// 
  /// Editors should always interact with this value to ensure the proper
  /// transforms are applied.
  dynamic
    get value {
      final value = field.value;
      if (field.transform != null) {
        return field.transform(value);
      } else if (ValueTransform.canTransform(field.type, valueType)) {
        return ValueTransform.transform(value, field.type, valueType);
      } else {
        return value.toString();
      }
    }
    set value(value) {
      assert(isEnabled);
      if (field.transform != null && field.transform.hasInverse) {
        field.value = field.transform.inverse(value);
      } else {
        field.value = ValueTransform.transform(value, valueType, field.type);
      }
    }
    
  factory FieldEditor.forField(Field field) {
    assert(factory != null);
    return factory(field);
  }
    
  FieldEditor(this.field) {
    field.onValueChange.listen((_) => refreshElement());
    scheduleMicrotask(refreshElement);
  }

  /// Refresh this field editor's [element].
  /// 
  /// This method is called whenever this editor's [value] or [isEnabled] state
  /// changes.
  void refreshElement();
}
