part of fields.view;

final Tag objectEditor = createTag((_) => new ObjectEditor());

class ObjectEditorResources extends Resources {
  String get style =>
"""
:host {
  background-color: ${backgroundColor};
  border:1px solid ${borderColor};
  height:100%;    
  overflow:auto;
  width:100%;
}
button, input {
  background-color:transparent;
  border-style:none;
  color:${textColor};
}
input[type="text"] {  
  width:100%;
}
button:disabled, input[type="text"][disabled] { 
  color:${textDisabledColor};
}
input[class=error] {
  border: 2px solid ${errorColor};
  margin: 0;
}
select {
  background-color:transparent;
  border: 0;
  border-radius: 0;
  color:${textColor};
}
optgroup, option {
  background: ${fieldBackgroundColor};
}
table {
  border-collapse:collapse;
  border-top:1px solid ${borderColor};
  border-left:1px solid ${borderColor};
  margin:0px;            
  padding:0px;
  table-layout:fixed;
  width:100%;
}
td {
  background-color: ${fieldBackgroundColor};
  border-top:1px solid ${borderColor};
  border-left:1px solid ${borderColor};
  color: ${textColor};
  vertical-align:text-top;
  margin:0px;
  overflow:hidden;
  padding:0px;
  text-overflow:ellipsis;
}
.expander {
  background-color:transparent;
  border-style:none;
  color:${textColor};
  float:left;   
  width:16px;
  padding:0;
  margin-top:4px;
}
.editorbox {
  display:block; 
  overflow:hidden;  
  text-overflow:ellipsis;  
  white-space:nowrap;
}
.filterbox {  
  display:block; 
  overflow:hidden;
}
.filter {
  background-color:transparent;
  border-style:none;
  color:${textColor};
  width:100%;
}
.namebox {
  padding-left:4px;
  width:20%;
}
.toggle {
  background-color:transparent;
  border-style:none;
  color:${textColor};
  float:left;
}
.toggle:disabled {
  opacity:0.4;
}
.toggle[down] {
  background-color: ${fieldBackgroundColor};
}
""";
  final String backgroundColor      = 'transparent';
  final String borderColor          = 'grey';
  final String errorColor           = 'red';
  final String fieldBackgroundColor = 'transparent';
  final String textColor            = 'black';
  final String textDisabledColor    = 'grey';
  const ObjectEditorResources();
}

class ObjectEditorDarkResources extends ObjectEditorResources {
  final String backgroundColor      = '#2d2d30';
  final String borderColor          = '#2d2d30';
  final String fieldBackgroundColor = '#252526';
  final String textColor            = '#f1f1f1';
  const ObjectEditorDarkResources();
}

class ObjectEditor extends EditorView<Object> with Themeable {
  
  static const TAG = 'object-editor';
  static ObjectEditorResources _factoryResources;

  String _filterText = '';  
  Function _filter;
  Function _filterPrivate;
    
  final ObjectEditorResources resources;
  final StyleElement styleElement = new StyleElement();
    
  ButtonElement _sortButton;
  ButtonElement _privateButton;
  InputElement _filterBox;
  
  FieldList _list;
  FieldListView _listView;
  StreamSubscription _listListener;
  
  final _onFieldChange = new StreamController<Field>.broadcast();
  Stream<Field> get onFieldChange => _onFieldChange.stream;
  
  /// Creates a new object editor with the given initial [selection] and 
  /// [resources].
  factory ObjectEditor({Object selection,
      ObjectEditorResources resources: const ObjectEditorResources()}) {    
    _factoryResources = resources;
    final editor =  new Element.tag('div', TAG);
    editor.initialize(selection);
    return editor;
  }
  
  ObjectEditor.created() 
  : super.created()
  , resources = _factoryResources;
  
  void initialize_(Object selection, _) {
    
    _filter = (Field f) => f.name.contains(_filterText);
    _filterPrivate = (Field f) => !f.isPrivate && _filter(f);
    
    children(this, () {
      shadow(() {
        append(styleElement)..text = resources.style;
        _sortButton = toggleButton(() {
          span()..innerHtml = '&#9650;'..classes = ['up'];
          span()..innerHtml = '&#9660;'..classes = ['down'];
        })
        ..classes = ['toggle']
        ..disabled = true
        ..onToggle.listen((bool isDown) {
          if (isDown) {
            _list.comparator = 
                (Field a, Field b) => a.name.compareTo(b.name) * -1;
          } else {
            _list.comparator = FieldList.defaultComparator;
          }
          _refresh();
        });
        _privateButton = toggleButton(() {
          span()..text = '_'..classes = ['up down'];
        })
        ..classes = ['toggle']
        ..disabled = true
        ..onToggle.listen((bool isDown) {
          if (isDown) {
            _list.predicate = _filter;
          } else {
            _list.predicate = _filterPrivate;
          }
          _refresh();
        });
        span(() {
          _filterBox = input()
          ..type = 'text'
          ..classes = ['filter']
          ..disabled = true
          ..placeholder = 'filter fields';
        })
        ..classes = ['filterbox']
        ..onInput.listen((_) {
          _filterText = _filterBox.value;
          _list.refreshView();
          _refresh();
        });
        _listView = append(new FieldListView());
      });
    });
  }
  
  /// Refreshes this object editor to reflect the current state of the
  /// [selection] object.
  /// 
  /// This method does _not_ need to be called for any changes made to the 
  /// [selection] object through this object editor or any field editors.  It 
  /// _does_ need to be called when the [selection] object is modified external
  /// to this library, for example by another user interface element.
  void refresh() {
    if (hasSelection) {
      _list.refreshSource(selection);
      _refresh();
    }
  }
  
  void _refresh() { 
    _sortButton.disabled = !hasSelection;
    _privateButton.disabled = !hasSelection;
    _filterBox.disabled = !hasSelection;
    _listView.refresh(_list);
  }
  
  void select_(Object value) {
    if (_listListener != null) _listListener.cancel();
    _list = new FieldList()
    ..refreshSource(value)
    ..predicate = _filterPrivate;      
    _listListener = _list.onFieldChange.listen((f) {
      _onFieldChange.add(f);
    });
    _listView.clear();
    _refresh();
  }
}
